/*
 * image.c
 *
 * Copyright (c) 1995 Erik Corry ehcorry@inet.uni-c.dk
 * Copyright (c) 1995 John Kilburg (john@isri.unlv.edu)
 *
 * See the file COPYRIGHT for details.
 */
#include <stdlib.h>
#include <stdio.h>
#include <X11/Intrinsic.h>

#include "colorcube.h"
#include "dispdither.h"
#include "xcolorcube.h"
#include "image.h"
#include "endian.h"
#include "imagep.h"
#include "image_format.h"


typedef void (*lt_expansion_fn)();

struct ccs_image_pstate
{
  Display     *dpy;
  Colormap    cmap;
  int         depth;
  Visual      *v;
  XColor      bgcolor;
  XColor      fgcolor;

  cct_cube    colorcube;            /* cuboid of allocated colors on screen */
  cct_cube    grayscale;            /* gray shades allocated on screen */
  cct_dither_table color_tables[4]; /* color dither table */
  cct_dither_table gray_tables[4];  /* color dither table */
  struct ccs_special_color special_colors[256]; /* specially allocated colors */
  int         special_count;
  int         special_max;
};

struct ccs_image_state
{
  struct ccs_image_pstate *ips;

  int format;
  ImageLineProc linecallback;
  void *lineclosure;
  
  cct_dither_table dither_table[4];  /* Table to dither/convert to screen */
  int free_dither_table;             /* How many tables were allocated (0-4) */
  ddt_dither_fn dither_function;     /* Function to dither/convert with */
  bool dithering;
  bool data_ended;

  XImage *xi;                        /* X image */

  struct ifs_vector if_vector;       /* Image format decoder info */

  Pixel bg, fg;                      /* background, foreground for bitmaps */

  unsigned char *expansion_buf;      /* Area for upgrading to bigger depth */
  lt_expansion_fn expansion_function;/* Function for upgrading to bigger depth*/

  int expanded_type;            /* Image type after expansion */
  int image_special_count;      /* colors allocated for this image */
  int image_special_max;        /* maximum colors to allocate for this image */
  bool no_more_specials;
  bool special_blacklist[256];  /* Have we already tried to allocate this pixel once */
  bool special_real_blacklist[256];  /* Have we already really tried to allocate this pixel once */
};

static void ImageToXImage(void *pointer, int fline, int lline);

static unsigned char *
lf_get_8_bit_map(Intensity *old_table)
{
  unsigned char *answer = calloc_mem(1, 256);
  int i;

  for (i = 0; i < 256; i++)
    answer[i] = old_table[i] >> 8;

  return answer;
}

static bool
lf_null_gray_map(unsigned char *maps[3])
{
  int i;
  for (i = 0; i < 256; i++)
  {
    if (maps[0][i] != maps[1][i]) return false;
    if (maps[2][i] != maps[1][i]) return false;
    if (maps[2][i] != i) return false;
  }
  return true;
}

static Image *
lf_get_image(struct ccs_image_state *is)
{
  if (is->if_vector.getImageProc && is->if_vector.image_format_closure)
    return(is->if_vector.getImageProc(is->if_vector.image_format_closure));
  return(NULL);
}

static void
lf_set_up_bitmap(struct ccs_image_state *is)
{
  Image *image = lf_get_image(is);
  struct ccs_image_pstate *ips = is->ips;
  bool really_allocated = false;
  int i;
  Intensity intensities[3];

  is->expanded_type = image->type;
  
  if(image->type == IBITMAP)
  {
    is->fg = ips->fgcolor.pixel;
    is->bg = ips->bgcolor.pixel;
    return;
  }

  for(i = 0; i < 2; i++)
  {
    bool really_allocated_last_time = really_allocated;
    if(image->transparent != i)
    {
      intensities[0] = image->rgb.red[i];
      intensities[1] = image->rgb.green[i];
      intensities[2] = image->rgb.blue[i];
      if (!xccf_allocate_special(ips->dpy,
				 ips->cmap,
				 ips->colorcube,
				 ips->grayscale,
				 intensities,
				 &really_allocated,
				 /* don't-really-allocate-flag */
				 ips->special_count >= ips->special_max,
				 ips->special_colors,
				 ips->special_count,
				 i ? &(is->fg): &(is->bg)))
      {
        if(really_allocated_last_time)
        {
          XFreeColors(ips->dpy,
		      ips->cmap,
		      &(ips->special_colors[ips->special_count - 1].pixel),
		      1,
		      0);
         ips->special_count--;
        }

	if(image->rgb.red[0] == image->rgb.green[0] &&
	   image->rgb.red[0] == image->rgb.blue[0] &&
	   image->rgb.red[1] == image->rgb.green[1] &&
	   image->rgb.red[1] == image->rgb.blue[1])
        {
          is->expanded_type = IGRAY;
	  is->expansion_buf = calloc_mem(1, image->width + 64);
        }
        else
        {
	  is->expanded_type = ITRUE;
	  is->expansion_buf = calloc_mem(1, (3 * image->width) + 192);
        }

	return;
      }
      if(really_allocated) ips->special_count++;
    }
    else
    {
      if (i) is->fg = ips->bgcolor.pixel;
      else is->bg = ips->bgcolor.pixel;
    }
  }

  return;
}

static void
lf_expand_bitmap_to_true(
  byte *expansion_buffer,
  byte *source_buffer,
  int width,
  Intensity *red_cmap,
  Intensity *green_cmap,
  Intensity *blue_cmap,
  int transparent,
  Intensity red_bg,
  Intensity green_bg,
  Intensity blue_bg)
{
  int i;
  byte red[2];
  byte green[2];
  byte blue[2];
  red[0] = red_cmap[0] >> 8;
  red[1] = red_cmap[1] >> 8;
  green[0] = green_cmap[0] >> 8;
  green[1] = green_cmap[1] >> 8;
  blue[0] = blue_cmap[0] >> 8;
  blue[1] = blue_cmap[1] >> 8;
  if(transparent != -1)
  {
    red[transparent] = red_bg >> 8;
    green[transparent] = green_bg >> 8;
    blue[transparent] = blue_bg >> 8;
  }
  for(i = (width + CHAR_BITS - 1) / CHAR_BITS; i; i--)
  {
    byte t = *source_buffer++;
#   ifdef LITTLE_ENDIAN
      expansion_buffer[0] =   red[t & 1];
      expansion_buffer[1] = green[t & 1];
      expansion_buffer[2] =  blue[t & 1];
      expansion_buffer[3] =   red[t >> 1 & 1];
      expansion_buffer[4] = green[t >> 1 & 1];
      expansion_buffer[5] =  blue[t >> 1 & 1];
      expansion_buffer[6] =   red[t >> 2 & 1];
      expansion_buffer[7] = green[t >> 2 & 1];
      expansion_buffer[8] =  blue[t >> 2 & 1];
      expansion_buffer[9] =   red[t >> 3 & 1];
      expansion_buffer[10] = green[t >> 3 & 1];
      expansion_buffer[11] =  blue[t >> 3 & 1];
      expansion_buffer[12] =   red[t >> 4 & 1];
      expansion_buffer[13] = green[t >> 4 & 1];
      expansion_buffer[14] =  blue[t >> 4 & 1];
      expansion_buffer[15] =   red[t >> 5 & 1];
      expansion_buffer[16] = green[t >> 5 & 1];
      expansion_buffer[17] =  blue[t >> 5 & 1];
      expansion_buffer[18] =   red[t >> 6 & 1];
      expansion_buffer[19] = green[t >> 6 & 1];
      expansion_buffer[20] =  blue[t >> 6 & 1];
      expansion_buffer[21] =   red[t >> 7 & 1];
      expansion_buffer[22] = green[t >> 7 & 1];
      expansion_buffer[23] =  blue[t >> 7 & 1];
#   else
      expansion_buffer[0] =   red[t >> 7 & 1];
      expansion_buffer[1] = green[t >> 7 & 1];
      expansion_buffer[2] =  blue[t >> 7 & 1];
      expansion_buffer[3] =   red[t >> 6 & 1];
      expansion_buffer[4] = green[t >> 6 & 1];
      expansion_buffer[5] =  blue[t >> 6 & 1];
      expansion_buffer[6] =   red[t >> 5 & 1];
      expansion_buffer[7] = green[t >> 5 & 1];
      expansion_buffer[8] =  blue[t >> 5 & 1];
      expansion_buffer[9] =   red[t >> 4 & 1];
      expansion_buffer[10] = green[t >> 4 & 1];
      expansion_buffer[11] =  blue[t >> 4 & 1];
      expansion_buffer[12] =   red[t >> 3 & 1];
      expansion_buffer[13] = green[t >> 3 & 1];
      expansion_buffer[14] =  blue[t >> 3 & 1];
      expansion_buffer[15] =   red[t >> 2 & 1];
      expansion_buffer[16] = green[t >> 2 & 1];
      expansion_buffer[17] =  blue[t >> 2 & 1];
      expansion_buffer[18] =   red[t >> 1 & 1];
      expansion_buffer[19] = green[t >> 1 & 1];
      expansion_buffer[20] =  blue[t >> 1 & 1];
      expansion_buffer[21] =   red[t & 1];
      expansion_buffer[22] = green[t & 1];
      expansion_buffer[23] =  blue[t & 1];
#   endif
    expansion_buffer += 24;
  }
}

static void
lf_expand_bitmap_to_gray(
  byte *expansion_buffer,
  byte *source_buffer,
  int width,
  Intensity *red_cmap,
  Intensity *green_cmap,
  Intensity *blue_cmap,
  int transparent,
  Intensity red_bg,
  Intensity green_bg,
  Intensity blue_bg)
{
  int i;
  byte gray[2];
  gray[0] = red_cmap[0] >> 8;
  gray[1] = red_cmap[1] >> 8;
  if(transparent != -1)
    gray[transparent] = red_bg >> 8;

  for(i = (width + CHAR_BITS - 1) / CHAR_BITS; i; i--)
  {
    byte t = *source_buffer++;
#   ifdef LITTLE_ENDIAN
      expansion_buffer[0] =   gray[t & 1];
      expansion_buffer[1] =   gray[t >> 1 & 1];
      expansion_buffer[2] =   gray[t >> 2 & 1];
      expansion_buffer[3] =   gray[t >> 3 & 1];
      expansion_buffer[4] =   gray[t >> 4 & 1];
      expansion_buffer[5] =   gray[t >> 5 & 1];
      expansion_buffer[6] =   gray[t >> 6 & 1];
      expansion_buffer[7] =   gray[t >> 7 & 1];
#   else
      expansion_buffer[0] =   gray[t >> 7 & 1];
      expansion_buffer[1] =   gray[t >> 6 & 1];
      expansion_buffer[2] =   gray[t >> 5 & 1];
      expansion_buffer[3] =   gray[t >> 4 & 1];
      expansion_buffer[4] =   gray[t >> 3 & 1];
      expansion_buffer[5] =   gray[t >> 2 & 1];
      expansion_buffer[6] =   gray[t >> 1 & 1];
      expansion_buffer[7] =   gray[t & 1];
#   endif
    expansion_buffer += 8;
  }
}


static void
lf_set_up_expansion(struct ccs_image_state *is)
{
  if(is->expanded_type == IGRAY)
    is->expansion_function = &lf_expand_bitmap_to_gray;
  else
    is->expansion_function = &lf_expand_bitmap_to_true;
}

static void
lf_set_up_dither(struct ccs_image_state *is)
{
  unsigned char *maps[3];
  int i;
  cct_8_true_conversion_table cdt;
  struct ccs_image_pstate *ips = is->ips;
  Image *image = lf_get_image(is);
  int image_type;

  is->dithering = false;

  if(is->expansion_buf)
  {
    if(image->depth > 1)
      is->expanded_type = image_type = ITRUE;
    else
      image_type = is->expanded_type; /* lf_set_up_bitmap has already set it */
  }
  else
  {
    image_type = image->type;
  }

  maps[0] = maps[1] = maps[2] = 0;

  if (image->type == IRGB || image->type == IGRAY)
  {
    maps[0] = lf_get_8_bit_map(image->rgb.red);
    maps[1] = lf_get_8_bit_map(image->rgb.green);
    maps[2] = lf_get_8_bit_map(image->rgb.blue);
  }

  if (image->transparent != -1)
  {
    maps[0][image->transparent] = ips->bgcolor.red >> 8;
    maps[1][image->transparent] = ips->bgcolor.green >> 8;
    maps[2][image->transparent] = ips->bgcolor.blue >> 8;    
  }

  /*
   * A real color dither/conversion?
   */
  if ((image_type == IRGB || image_type == ITRUE) && ips->colorcube)
  {
    if (image_type == IRGB &&
	ips->colorcube->cube_type == cube_true_color)
    {
      /*
       * colormap to true-color conversion needed
       */
      if (!ips->color_tables[0].true_true_conversion)
      {
	ips->color_tables[0].true_true_conversion =
	    ccf_create_true_true_conversion_table(ips->colorcube);
      }
      cdt = ccf_true_true_to_8_true_conversion_table(
						     ips->color_tables[0].
						     true_true_conversion,
						     image->rgb.used,
						     maps);
      is->dither_table[0].eight_true_conversion = cdt;
      is->dither_table[1].eight_true_conversion = cdt;
      is->dither_table[2].eight_true_conversion = cdt;
      is->dither_table[3].eight_true_conversion = cdt;
      is->free_dither_table = 1;
      is->dither_function = &ddf_convert_line_8;
      is->dithering = false;
    }
    else if (image_type == IRGB &&
	     (ips->colorcube->cube_type == cube_mapping ||
	      ips->colorcube->cube_type == cube_no_mapping))
    {
      /*
       * 8-bit palette to 8-bit screen dither
       */
      if (!ips->color_tables[0].true_8_dither)
	  for (i = 0; i < 4; i++)
	      ips->color_tables[i].true_8_dither =
		  ccf_create_true_8_dither_table(i, 4, 4, ips->colorcube);
      for (i = 0; i < 4; i++)
      {
	is->dither_table[i].eight_8_dither =
	    ccf_true_8_to_8_8_dither_table(ips->color_tables[i].true_8_dither,
					   image->rgb.used, maps);
        if(image->transparent != -1)
	    ccf_set_specially_allocated(is->dither_table[i].eight_8_dither,
					image->transparent,
					ips->bgcolor.pixel);
      }
      is->free_dither_table = 4;
      is->dither_function = &ddf_dither_line_8;
      is->dithering = true;
    }
    else if (image_type == ITRUE &&
	     ips->colorcube->cube_type == cube_true_color)
    {
      /*
       * true-true conversion job
       */
      if (!ips->color_tables[0].true_true_conversion)
	  ips->color_tables[0].true_true_conversion =
	      ccf_create_true_true_conversion_table(ips->colorcube);
      for (i = 0; i < 4; i++)
	  is->dither_table[i].true_true_conversion =
	      ips->color_tables[0].true_true_conversion;
      is->free_dither_table = 0; /* don't free it its a copied pointer!! */
      is->dither_function = &ddf_convert_line_24;
      is->dithering = false;
    }
    else if (image_type == ITRUE &&
	     (ips->colorcube->cube_type == cube_mapping ||
	      ips->colorcube->cube_type == cube_no_mapping))
    {
      /*
       * true-color to 8-bit dither-type-situation
       */
      if (!ips->color_tables[0].true_8_dither)
	  for (i = 0; i < 4; i++)
	      ips->color_tables[i].true_8_dither =
		  ccf_create_true_8_dither_table(i, 4, 4, ips->colorcube);
      for (i = 0; i < 4; i++)
	  is->dither_table[i].true_8_dither =
	      ips->color_tables[i].true_8_dither;
      is->free_dither_table = 0; /* don't free it its a copied pointer!! */
      is->dither_function = &ddf_color_dither_line_24;
      is->dithering = true;
    }
  }
  /*
   * else a black and white dither
   */
  else if ((image_type == IRGB || image_type == ITRUE ||
	    image_type == IGRAY) &&
	   ips->grayscale && ips->grayscale->u.grayscale.value_count < 29)
  {
    if (!ips->gray_tables[0].eight_8_dither)
	for (i = 0; i < 4; i++)
	    ips->gray_tables[i].eight_8_dither =
		ccf_create_gray_dither_table(i, 4, 4, ips->grayscale);
    if (image_type == IRGB || image_type == IGRAY)
    {
      /*
       * colormap to grayscale dither
       */
      if (lf_null_gray_map(maps) && image->transparent == -1)
      {
	for (i = 0; i < 4; i++)
	    is->dither_table[i].eight_8_dither =
		ips->gray_tables[i].eight_8_dither;
	is->free_dither_table = 0; /* don't free it its a copied pointer */
      }
      else
      {
	for (i = 0; i < 4; i++)
	{
	  is->dither_table[i].eight_8_dither =
	      ccf_gray_to_gray_dither_convert(ips->gray_tables[i].
					      eight_8_dither,
					      image->rgb.used,
					      maps);
          if(image->transparent != -1)
	      ccf_set_specially_allocated(is->dither_table[i].eight_8_dither,
					  image->transparent,
					  ips->bgcolor.pixel);

	}
	is->free_dither_table = 4;
      }
      is->dither_function = &ddf_dither_line_8;
      is->dithering = true;
    }
    else
    {
      /*
       * True-color-to-grayscale type situtation
       */
      for (i = 0; i < 4; i++)
	  is->dither_table[i].true_8_dither = 
	      ips->gray_tables[i].true_8_dither;
      is->free_dither_table = 0; /* don't free it its a copied pointer!! */
      is->dither_function = &ddf_gray_dither_line_24;
      is->dithering = true;
    }
  }
  /*
   * else a bw conversion
   */
  else if ((image_type == IRGB || image_type == ITRUE ||
	    image_type == IGRAY) &&
	   ips->grayscale &&
	   ips->grayscale->u.grayscale.value_count >= 29)
  {
    if (image->transparent != -1 &&
	(image_type == IRGB || image_type == IGRAY))
    {
      maps[0][image->transparent] = ips->bgcolor.red >> 8;
      maps[1][image->transparent] = ips->bgcolor.green >> 8;
      maps[2][image->transparent] = ips->bgcolor.blue >> 8;
    }
    
    if (!ips->gray_tables[0].eight_true_conversion)
	ips->gray_tables[0].eight_true_conversion =
	    ccf_create_gray_conversion_table(ips->grayscale);
    if (image_type == IGRAY || image_type == IRGB)
    {
      /*
       * Simple remap of input values to output values
       */
      if (lf_null_gray_map(maps))
      {
	for (i = 0; i < 4; i++)
	    is->dither_table[i].eight_true_conversion =
		ips->gray_tables[0].eight_true_conversion;
	is->free_dither_table = 0; /* don't free it its a copied pointer */
      }
      else
      {
	for (i = 0; i < 4; i++)
	    is->dither_table[i].eight_true_conversion =
		ccf_gray_to_gray_conversion_convert(
						ips->gray_tables[0].
						  eight_true_conversion, 
						image->rgb.used,
						maps);
	is->free_dither_table = 4;
      }
      is->dither_function = &ddf_convert_line_8;
      is->dithering = false;
    }
    else
    {
      /*
       * Reduce 24-bit color to grayscale without dithering
       */
      for (i = 0; i < 4; i++)
	    is->dither_table[i].eight_true_conversion =
		ips->gray_tables[0].eight_true_conversion;
      is->free_dither_table = 0; /* don't free it its a copied pointer */
      is->dither_function = &ddf_gray_convert_line_24;
      is->dithering = true;
    }
  }
  
  if(maps[0]) free_mem(maps[0]);
  if(maps[1]) free_mem(maps[1]);
  if(maps[2]) free_mem(maps[2]);
  maps[0] = maps[1] = maps[2] = 0;

  return;
}


/*
 * I'm going to need this soon:

       *special_return =
       cube_table.true_true_conversion->pixel_values[0][intensities[0] >> 8] +
       cube_table.true_true_conversion->pixel_values[1][intensities[1] >> 8] +
       cube_table.true_true_conversion->pixel_values[2][intensities[2] >> 8];
 *
 */


static void
lf_get_new_specials(
  struct ccs_image_state *is,
  struct ccs_image_pstate *ips,
  Image *image,
  byte *input)
{
  bool *blacklist = is->special_blacklist;
  bool *real_blacklist = is->special_real_blacklist;
  /*
   * Find 3 identical pixels and win a new special color allocation. The
   * pixels are searched for spaced 4 apart, because this ensures that
   * they are not just antialias pixels on the edge of writing or other
   * unimportantly strewn pixels. Also, this catches the case where the
   * GIF has already beed dithered with a 4x4 matrix. This is an important
   * case because it can interfere with our own dithering and produce
   * ugly artefacts.
   */
  if (!is->expansion_buf &&
      is->dithering &&
      image->type == IRGB &&
      !is->no_more_specials)
  {
    int i;
    for(i = 0; i < image->width - 8; i++, input++)
    {
      byte t = *input;
      bool dont_really_allocate;
      if (real_blacklist[t])
      {
        continue;
      }

      if (ips->special_count < ips->special_max &&
	  is->image_special_count < is->image_special_max &&
          t == input[4] &&
          t == input[8])
      {
        real_blacklist[t] = true;
        dont_really_allocate = false;
      }
      else
      {
        if(blacklist[t]) continue;
        dont_really_allocate = true;
      }
      blacklist[t] = true;
      {
	Intensity intensities[3];
	bool really_allocated;
	unsigned long special_entry;

	if(t == image->transparent)
	  continue;

	intensities[0] = image->rgb.red[t];
	intensities[1] = image->rgb.green[t];
	intensities[2] = image->rgb.blue[t];
	if (!xccf_allocate_special(ips->dpy,
				   ips->cmap,
				   ips->colorcube,
				   ips->grayscale,
				   intensities,
				   &really_allocated,
				   dont_really_allocate,
				   ips->special_colors,
				   ips->special_count,
				   &special_entry))
	{
	  /*
	   * If a color allocation fails once, we won't bother the
	   * X server again. Color allocation is a slow round trip
	   */
	  if (!dont_really_allocate)
	    is->image_special_max = is->image_special_count;
	}
	else
	{
	  int j;
	  if(really_allocated)
	  {
	    ips->special_count++;
	    is->image_special_count++;
	    if(is->image_special_count >= image->rgb.used)
	      is->no_more_specials = true;
	    for(j = 0; j < 256; j++)
	      blacklist[j] = false;
	  }
          for(j = 0; j < 4; j++)
          {
	    ccf_set_specially_allocated(is->dither_table[j].eight_8_dither,
	                                t,
	                                special_entry);
	  }
	}
      }
    }
  }
}

static void
ImageToXImage(void *pointer, int fline, int lline)
{
  struct ccs_image_state *is = (struct ccs_image_state *)pointer;
  struct ccs_image_pstate *ips = is->ips;
  unsigned char *input;
  unsigned char *output;
  int depth;
  int line;
  Image *image = lf_get_image(is);
  int image_type = image->type;

  /*
   * First time
   */
  if (is->xi == NULL)
  {
    /*
     * No image gets more than half the special colors
     */
    is->image_special_max = (ips->special_max - ips->special_count) / 2;

    if (image->pixlen == 1) depth = 1;
    else depth = ips->depth;
    
    if(depth == 1)
    {
      lf_set_up_bitmap(is);
      if(is->expansion_buf)
        lf_set_up_expansion(is);
      image_type = is->expanded_type;
    }
    else
    {
      is->expanded_type = image->type;
    }

    if (is->expansion_buf) depth = ips->depth;

    if (image_type == IRGB || image_type == ITRUE)
    {
      if (!ips->colorcube)
      {
        ips->colorcube =
	    xccf_allocate_cube(ips->dpy, ips->cmap, ips->v, ips->depth);
      }
    }
    if ((image_type == IGRAY || !ips->colorcube) && image->pixlen != 1)
    {
      if (!ips->grayscale)
      {
	ips->grayscale =
	    xccf_allocate_grays(ips->dpy, ips->cmap, ips->v, ips->depth);
      }
    }
    
    is->xi = XCreateImage(ips->dpy, ips->v,
			  depth,
                          depth == 1 ? XYBitmap : ZPixmap,
                          0, NULL,
			  image->width, image->height,
			  32, 0);
    
    /* Make sure we have plenty of padding at the end of each line */
    is->xi->bytes_per_line += 16;
    is->xi->data = (char *)calloc_mem(1,
				      image->height *
				      is->xi->bytes_per_line);
    
#ifdef LITTLE_ENDIAN
    is->xi->byte_order = is->xi->bitmap_bit_order = LSBFirst;
#else
    is->xi->byte_order = is->xi->bitmap_bit_order = MSBFirst;
#endif

    if(depth != 1)
    {
      XAddPixel(is->xi, ips->bgcolor.pixel);
    }
    if(depth != 1 || is->expansion_buf)
    {
      lf_set_up_dither(is);
    }
  }

  /*
   * Every time
   */

  input = image->data + image->bytes_per_line * fline;
  output = is->xi->data + is->xi->bytes_per_line * fline;

  lf_get_new_specials(is, ips, image, input);

  for (line = fline;
       line <= lline;
       line++, input += image->bytes_per_line, output += is->xi->bytes_per_line)
  {
    if (image->pixlen == 1 && !is->expansion_buf)
    {
      memcpy(output, input, MIN(is->xi->bytes_per_line, image->bytes_per_line));
    }
    else
    {
      byte *expanded_input;
      if (is->expansion_buf)
      {
        is->expansion_function(is->expansion_buf,
                               input,
                               image->width,
                               image->rgb.red,
                               image->rgb.green,
                               image->rgb.blue,
                               image->transparent,
                               is->ips->bgcolor.red,
                               is->ips->bgcolor.green,
                               is->ips->bgcolor.blue);
        expanded_input = is->expansion_buf;
      }
      else
      {
        expanded_input = input;
      }
      is->dither_function(is->dither_table[line & 3],
			  expanded_input,
			  output,
			  is->xi->bits_per_pixel,
			  image->width);
    }
  }
  
  if (is->linecallback != NULL)
  {
    (is->linecallback)(is->xi, fline, lline, is->lineclosure);
  }
  
  return;
}

static int lv_special_max = -1;

/*
 * ImagePState
 */
ImagePState
ImagePInit(dpy, v, depth, cmap, background, foreground)
Display *dpy;
Visual *v;
int depth;
Colormap cmap;
Pixel background;
Pixel foreground;
{
  struct ccs_image_pstate *ips;

  ips = (struct ccs_image_pstate *)alloc_mem(sizeof(struct ccs_image_pstate));
  memset(ips, 0, sizeof(struct ccs_image_pstate));

  ips->dpy = dpy;
  ips->depth = depth;
  ips->v = v;
  ips->cmap = cmap;

  ips->bgcolor.pixel = background;
  ips->bgcolor.flags = DoRed | DoGreen | DoBlue;
  XQueryColor(dpy, cmap, &(ips->bgcolor));

  ips->fgcolor.pixel = foreground;
  ips->fgcolor.flags = DoRed | DoGreen | DoBlue;
  XQueryColor(dpy, cmap, &(ips->fgcolor));

  if(lv_special_max == -1)
  {
    if(getenv("CHIMERA_MAX_SPECIAL_COLORS"))
      lv_special_max = atoi(getenv("CHIMERA_MAX_SPECIAL_COLORS"));
    else
      lv_special_max = 64;
  }

  ips->special_max = lv_special_max;

  return(ips);
}

/*
 * ImagePDestroy
 */
void
ImagePDestroy(ips)
ImagePState ips;
{
  struct ccs_image_pstate *ipsp = (struct ccs_image_pstate *)ips;

  free_mem(ipsp);

  return;
}

/*
 * ImageInit
 */
ImageState
ImageInit(ips, format, linecallback, lineclosure)
ImagePState ips;
int format;
ImageLineProc linecallback;
void *lineclosure;
{
  struct ccs_image_state *is;
  struct ccs_image_pstate *ipsp = (struct ccs_image_pstate *)ips;

  is = (struct ccs_image_state *)alloc_mem(sizeof(struct ccs_image_state));
  memset(is, 0, sizeof(struct ccs_image_state));

  is->ips = ipsp;
  is->format = format;
  is->linecallback = linecallback;
  is->lineclosure = lineclosure;
  is->fg = 0;
  is->bg = 0;

  is->if_vector.image_format = format;
  if (format == IMAGE_GIF) gifInit(ImageToXImage, is, &is->if_vector);
  else if (format == IMAGE_PNM) pnmInit(ImageToXImage, is, &is->if_vector);
  else if (format == IMAGE_XBM) xbmInit(ImageToXImage, is, &is->if_vector);
#ifdef HAVE_JPEG
  else if (format == IMAGE_JPEG) jpegInit(ImageToXImage, is, &is->if_vector);
#endif
  else
  {
    ImageDestroy(is);
    is = 0;
  }

  return(is);
}

/*
 * ImageDestroy
 */
void
ImageDestroy(pointer)
ImageState pointer;
{
  struct ccs_image_state *is = (struct ccs_image_state *)pointer;
  int i;

  /*
   * This is wrong! What I want to do is free the colors when
   * a new page is selected. At the moment this is the same
   * as when the first image on the page is destroyed, but...
   */

  if(is->ips->special_count)
  {
    int i;
    for(i = 0; i < is->ips->special_count; i++)
      XFreeColors(is->ips->dpy,
                  is->ips->cmap,
                  &is->ips->special_colors[i].pixel,
                  1, 0);
    is->ips->special_count = 0;
  }


  if (is->if_vector.destroyProc && is->if_vector.image_format_closure != NULL)
  {
    is->if_vector.destroyProc(is->if_vector.image_format_closure);
    is->if_vector.image_format_closure = NULL;
  }
  if(is->xi)
  {
    if(is->xi->data)
      free_mem(is->xi->data);
    is->xi->data = 0;
    XDestroyImage(is->xi);
    is->xi = 0;
  }
  for(i = 0; i < is->free_dither_table; i++)
  {
    if(is->dither_table[i].generic_dither_table)
      free_mem(is->dither_table[i].generic_dither_table);
    is->dither_table[i].generic_dither_table = 0;
  }
  for( ; i < 4; i++)
    is->dither_table[i].generic_dither_table = 0;
  is->free_dither_table = 0;
  if(is->expansion_buf) free_mem(is->expansion_buf);
  is->expansion_buf = 0;
  free_mem(is);

  return;
}

static int debug_add_data_procs = -1;

/*
 * ImageAddData
 */
int
ImageAddData(is, data, len)
ImageState is;
void *data;
int len;
{
  struct ccs_image_state *isp = (struct ccs_image_state *)is;

  if(debug_add_data_procs == -1)
    debug_add_data_procs = !!getenv("CHIMERA_DEBUG_ADD_DATA_PROCS");

  if (isp->if_vector.addDataProc && isp->if_vector.image_format_closure)
  {
    if(!debug_add_data_procs)
      isp->if_vector.addDataProc(isp->if_vector.image_format_closure, data, len, isp->data_ended);
    else
    {
      int i;
      int s = atoi(getenv("CHIMERA_DEBUG_ADD_DATA_PROCS"));
      for(i = s; i < len + s - 1; i += s)
      {
        int status = isp->if_vector.addDataProc(isp->if_vector.image_format_closure,
                                                data, MIN(i,len), isp->data_ended);
        if(status == -1)
        {
          printf("Image read failure\n");
          break;
        }
        if(status == 0)
          break;
      }
    }
  }
  else return(-1);

  return(0);
}

/*
 * ImageEndData
 */
void
ImageEndData(is)
ImageState is;
{
  struct ccs_image_state *isp = (struct ccs_image_state *)is;
  isp->data_ended = true;
  return;
}

/*
 * ImageGetXImage
 */
XImage *
ImageGetXImage(is)
ImageState is;
{
  struct ccs_image_state *isp = (struct ccs_image_state *)is;

  return(isp->xi);
}

/*
 * ImageForeground
 */
Pixel
ImageForeground(is)
ImageState is;
{
  struct ccs_image_state *isp = (struct ccs_image_state *)is;

  return(isp->fg);
}

/*
 * ImageBackground
 */
Pixel
ImageBackground(is)
ImageState is;
{
  struct ccs_image_state *isp = (struct ccs_image_state *)is;

  return(isp->bg);
}


