/*
 * doc.c
 *
 * Copyright (c) 1995, John Kilburg (john@isri.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>

#include <signal.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Scrollbar.h>

#include "WWWP.h"

#define DOC_NEW 0
#define DOC_BACK 1
#define DOC_RELOAD 2

#define DT_TEXT 0
#define DT_STREAM_EXTERNAL 1
#define DT_UNKNOWN 2
#define DT_EXT 3

/*
 *
 * Private functions
 *
 */
static void DestroyDocInfo _ArgProto((DocInfo *));
static DocInfo *CreateDocInfo _ArgProto((WWWWidget, URLParts *, int));
static void DisplayText _ArgProto((DocInfo *, int));
static Boolean ProcessDocument _ArgProto((DocInfo *, Boolean));
static int DisplayFileExternal _ArgProto((DocInfo *));
static int HandleDownloadDocument _ArgProto((Document *, void *, int));
static int HandleDocument _ArgProto((Document *, void *, int));
static void CallSaveCallback _ArgProto((DocInfo *));

/*
 * DisplayFileExternal
 */
static int
DisplayFileExternal(di)
DocInfo *di;
{
  char *content;
  char *cmdline;
  char filename[L_tmpnam + 1];
  FILE *fp;
  int rval;
  MIMECap *c;

  if ((content = MIMEGetFieldValue(di->d->mflist, "Content-type", 0)) == NULL)
  {
    return(-1);
  }

  if ((c = MIMEFindViewer(content)) == NULL) return(-1);

  if ((fp = fopen(tmpnam(filename), "w")) == NULL) return(-1);
  if (fwrite(di->d->data, 1, di->d->len, fp) < di->d->len)
  {
    fclose(fp);
    unlink(filename);
    return(-1);
  }
  fclose(fp);
  
  cmdline = CreateCommandLine(c->command, di->rw->www.path, filename);
  if ((rval = fork()) == 0)
  {
    signal(SIGPIPE, SIG_DFL);
    system(cmdline);
    unlink(filename);
    exit(0);
  }

  free_mem(cmdline);

  return(rval);
}

/*
 * DisplayText
 */
static void
DisplayText(di, plain)
DocInfo *di;
int plain;
{
  char *url;
  DocInfo *ti;
  WWWWidget rw = di->rw;

  XawScrollbarSetThumb(rw->www.vert_bar, 0.0, 0.0);
  XawScrollbarSetThumb(rw->www.horiz_bar, 0.0, 0.0);

  WWWEndLayout(rw);

  if (di->action == DOC_BACK)
  {
    ti = rw->www.dilist;
    rw->www.dilist = rw->www.dilist->next;    
    DestroyDocInfo(ti);

    ti = rw->www.dilist;
    rw->www.dilist = rw->www.dilist->next;    
    DestroyDocInfo(ti);
  }
  else if (di->action == DOC_RELOAD)
  {
    DocInfo *ti;
    
    ti = rw->www.dilist;
    rw->www.dilist = rw->www.dilist->next;
    
    DestroyDocInfo(ti);
  }
  else
  {
    if (rw->www.dilist != NULL && rw->www.dilist->d != NULL)
    {
      DestroyDocument(rw->www.dilist->d);
      rw->www.dilist->d = NULL;
    }
  }

  WWWStartLayout(rw, di->up, !di->ishtml);

  di->inlist = True;

  di->next = rw->www.dilist;
  rw->www.dilist = di;

  if (rw->www.rurl != NULL) free_mem(rw->www.rurl);
  rw->www.rurl = MakeURL(di->up, 0);

  url = MakeURL(di->up, 0);
  XtCallCallbackList((Widget)rw, rw->www.displayCallbacks, (XtPointer)url);
  free_mem(url);

  return;
}

/*
 * CallSaveCallback
 */
static void
CallSaveCallback(di)
DocInfo *di;
{
  WWWCallbackInfo wwwci;

  wwwci.data = di->d->data;
  wwwci.len = di->d->len;
  wwwci.content = MIMEGetFieldValue(di->d->mflist, "content-type", 0);
  
  XtCallCallbackList((Widget)di->rw, di->rw->www.saveCallbacks,
		     (XtPointer)&wwwci);

  return;
}

/*
 * ProcessDocument
 *
 * Returns False if the download should continue.
 */
static Boolean
ProcessDocument(di, download)
DocInfo *di;
Boolean download;
{
  char *value;
  Document *d = di->d;
  URLParts *up;
  DocInfo *ndi;
  int (*func)();

  if (download) func = HandleDownloadDocument;
  else func = HandleDocument;

  if ((value = MIMEGetFieldValue(d->mflist, "location", 0)) != NULL)
  {
    if ((up = ParseURL(value)) != NULL)
    {
      ndi = CreateDocInfo(di->rw, up, di->action);

      up->reload = di->rw->www.reload ? 1:0;
      LoadDocument((IOContext)di->rw, up, func, ndi);

      DestroyURLParts(up);
      DestroyDocInfo(di);

      return(True);
    }
  }

  if (!download)
  {
    if ((value = MIMEGetFieldValue(d->mflist, "content-type", 0)) != NULL)
    {
      if (strcasecmp(value, "text/html") == 0)
      {
	di->ishtml = True;
	di->type = DT_TEXT;
	DisplayText(di, 0);
      }
      else if (strncasecmp(value, "text/", 5) == 0) 
      {
	di->type = DT_TEXT;
	DisplayText(di, 1);
      }
      else di->type = DT_EXT;
    }
    else di->type = DT_EXT;
  }

  return(False);
}

/*
 * HandleDownloadDocument
 */
static int
HandleDownloadDocument(d, closure, status)
Document *d;
void *closure;
int status;
{
  DocInfo *di = (DocInfo *)closure;

  di->d = d;

  if (status == DS_DOWNLOADING) return(0);
  else if (status == DS_FINISHED)
  {
    if (ProcessDocument(di, True)) return(-1);
    CallSaveCallback(di);    
    DestroyDocInfo(di);
  }
  else
  {
    DestroyDocInfo(di);
    return(-1);
  }

  return(0);
}

/*
 * HandleDocument
 */
static int
HandleDocument(d, closure, status)
Document *d;
void *closure;
int status;
{
  DocInfo *di = (DocInfo *)closure;

  di->d = d;

  if (status == DS_NEEDS_AUTH)
  {
    if (FetchAuthInfo(di->rw, di->d, di->up))
    {
      LoadDocument((IOContext)(di->rw), di->up, HandleDocument, di);
    }
    else
    {
      GetUserAuthInfo(di->rw, di->up);
      if (!di->inlist) DestroyDocInfo(di);
    }
    return(-1);
  }
  else if (status == DS_DOWNLOADING || status == DS_FINISHED)
  {
    if (di->type == DT_UNKNOWN)
    {
      if (ProcessDocument(di, False)) return(-1);
    }
    if (di->type == DT_TEXT && d->len > 0)
    {
      MLAddData(di->rw->www.hs, d->data, d->len);
    }
    if (status == DS_FINISHED)
    {
      if (di->type == DT_TEXT) MLEndData(di->rw->www.hs);
      else if (DisplayFileExternal(di) == -1) CallSaveCallback(di);
    }
  }
  else
  {
    if (!di->inlist) DestroyDocInfo(di);
    return(-1);
  }

  return(0);
}

/*
 * DestroyDocInfo
 */
static void
DestroyDocInfo(di)
DocInfo *di;
{
  if (di->up != NULL) DestroyURLParts(di->up);
  if (di->d != NULL) DestroyDocument(di->d);
  if (di->s >= 0) IOEnd((IOContext)di->rw, di->s);
  free_mem(di);

  return;
}

/*
 * CreateDocInfo
 */
static DocInfo *
CreateDocInfo(rw, up, action)
WWWWidget rw;
URLParts *up;
int action;
{
  DocInfo *di;
  char *url;

  url = MakeURL(up, 0);
  XtCallCallbackList((Widget)rw, rw->www.downloadCallbacks, (XtPointer)url);
  free_mem(url);

  di = (DocInfo *)alloc_mem(sizeof(DocInfo));
  di->rw = rw;
  di->vert_shown = 0.0;
  di->vert_top = 0.0;
  di->horiz_shown = 0.0;
  di->horiz_top = 0.0;
  di->action = action;
  di->next = NULL;
  di->up = DupURLParts(up);
  di->s = -1;
  di->type = DT_UNKNOWN;
  di->elen = 0;
  di->ishtml = False;
  di->inlist = False;

  return(di);
}

/*
 *
 * Public functions
 *
 */

/*
 * DownloadDocument
 */
void
DownloadDocument(rw, up)
WWWWidget rw;
URLParts *up;
{
  DocInfo *di;

  IOStop(rw);
  rw->www.reload = False;

  up->reload = 0;
  di = CreateDocInfo(rw, up, DOC_NEW);

  LoadDocument((IOContext)rw, di->up, HandleDownloadDocument, di);

  return;
}

/*
 * NewDocument
 */
void
NewDocument(rw, up)
WWWWidget rw;
URLParts *up;
{
  DocInfo *di;

  IOStop(rw);
  rw->www.reload = False;
  up->reload = 0;
  di = CreateDocInfo(rw, up, DOC_NEW);

  LoadDocument((IOContext)rw, di->up, HandleDocument, di);

  return;
}

/*
 * ReloadDocument
 */
void
ReloadDocument(rw)
WWWWidget rw;
{
  DocInfo *di;
  URLParts *up;

  IOStop(rw);
  rw->www.reload = True;

  if (rw->www.dilist == NULL) return;

  up = rw->www.dilist->up;
  up->reload = 1;

  di = CreateDocInfo(rw, up, DOC_RELOAD);
  LoadDocument((IOContext)rw, di->up, HandleDocument, di);

  return;
}

/*
 * BackDocument
 */
void
BackDocument(rw)
WWWWidget rw;
{
  DocInfo *di;
  URLParts *up;

  IOStop(rw);
  rw->www.reload = False;

  if (rw->www.dilist == NULL) return;
  if (rw->www.dilist->next == NULL) return;

  up = rw->www.dilist->next->up;
  up->reload = 0;

  di = CreateDocInfo(rw, up, DOC_BACK);
  
  LoadDocument((IOContext)rw, di->up, HandleDocument, di);
  
  return;
}

/*
 * HomeDocument
 */
void
HomeDocument(rw)
WWWWidget rw;
{
  DocInfo *di, *ti;

  if (rw->www.dilist == NULL) return;

  for (di = rw->www.dilist; di->next != NULL; )
  {
    ti = di;
    di = di->next;
    DestroyDocInfo(ti);
  }

  rw->www.dilist = NULL;

  NewDocument(rw, di->up);

  DestroyDocInfo(di);

  return;
}


