/*
 * inline.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for more information.
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "WWWP.h"

typedef struct inline_state
{
  WWWWidget rw;                   /* always need this guy around */
  Document *d;                    /* the inline document */
  URLParts *up;                   /* the URL of the inline doc */
  RenderPart *next;               /* next image in same image list */
  RenderPart *mp;                 /* master image */
  ImageState is;                  /* image decoder/dither state */
  int border_width;               /* image border width */
  bool added;                     /* added yet */
} InlineState;

/*
 *
 * Private Functions
 *
 */

static int HandleInlineDocument _ArgProto((Document *, void *, int));
static RenderPart *CreateInlineState _ArgProto((WWWWidget, MLTag *,
						URLParts *));
static void AddImageRenderPart _ArgProto((WWWWidget, RenderPart *, int, int));
static void DestroyInlineState _ArgProto((InlineState *));
static void LineCallback _ArgProto((XImage *, int, int, RenderPart *));
static void RenderInline _ArgProto((WWWWidget, RenderPart *, XEvent *));
static void SetupInline _ArgProto((WWWWidget, RenderPart *, Region, int, int));
static void AddFakeRenderPart _ArgProto((WWWWidget));
static Boolean ProcessInlineDocument _ArgProto((RenderPart *));

/*
 * SetupInline
 */
static void
SetupInline(rw, rp, r, line_y, line_height)
WWWWidget rw;
RenderPart *rp;
Region r;
int line_y;
int line_height;
{
  rp->y = line_y + line_height - rp->height;
  rp->renderable = True;
  if (XRectInRegion(r, rp->x, rp->y, rp->width, rp->height))
  {
    RenderInline(rw, rp, NULL);
  }

  return;
}

/*
 * RenderInline
 *
 * Renders an inline image
 */
static void
RenderInline(rw, rp, xe)
WWWWidget rw;
RenderPart *rp;
XEvent *xe;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);
  Display *dpy = XtDisplay(rw->www.child);
  Window win = XtWindow(rw->www.child);
  GC gc = rw->www.image_gc;
  XImage *xi;
  int srcx, srcy;
  int dstx, dsty;
  int x, y;
  int bx, by;
  unsigned int width, height;
  InlineState *mci;
  ImageState is;

  if (ci->mp != NULL)
  {
    mci = (InlineState *)GetRenderState(ci->mp);
    is = mci->is;
  }
  else is = ci->is;

  if (is == 0) return;
  if ((xi = ImageGetXImage(is)) == NULL) return;

  if (rp->anchorPart != NULL)
  {
    XSetForeground(dpy, rw->www.image_gc, rw->www.anchorColor);
    XDrawRectangle(dpy, win, gc,
		   rp->x, rp->y,
		   rp->width, rp->height);
  }

  x = rp->x + ci->border_width;
  y = rp->y + ci->border_width;

  if (xe == NULL)
  {
    srcx = 0;
    srcy = 0;
    dstx = x;
    dsty = y;
    width = xi->width;
    height = xi->height;
  }
  else
  {
    dstx = x < xe->xexpose.x ? xe->xexpose.x:x;
    dsty = y < xe->xexpose.y ? xe->xexpose.y:y;

    srcx = dstx - x;
    srcy = dsty - y;

    bx = x + xi->width;
    if (xe->xexpose.x + xe->xexpose.width < bx)
    {
      bx = xe->xexpose.x + xe->xexpose.width;
    }
    width = bx - dstx;

    by = y + xi->height;
    if (xe->xexpose.y + xe->xexpose.height < by)
    {
      by = xe->xexpose.y + xe->xexpose.height;
    }
    height = by - dsty;
  }

  if (xi->depth == 1)
  {
    XSetForeground(dpy, gc, ImageForeground(is));
    XSetBackground(dpy, gc, ImageBackground(is));
  }

  XPutImage(dpy, win, gc, xi,
	    srcx, srcy,
	    dstx, dsty,
	    width, height);
  
  return;
}

/*
 * LineCallback
 *
 * fline is the first line to be rendered
 * lline is the last line to be rendered
 */
static void
LineCallback(xi, fline, lline, rp)
XImage *xi;
int fline, lline;
RenderPart *rp;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);
  WWWWidget rw = ci->rw;
  Display *dpy = XtDisplay(rw->www.child);
  Window win = XtWindow(rw->www.child);
  GC gc = rw->www.image_gc;
  InlineState *fci;
  RenderPart *tp;

  if (!ci->added) AddImageRenderPart(rw, rp, xi->width, xi->height);

  for (tp = rp, fci = ci; tp != NULL;
       fci = (InlineState *)GetRenderState(tp), tp = fci->next)
  {
    if (IsRenderable(tp))
    {
      if (tp->anchorPart != NULL)
      {
	XSetForeground(dpy, rw->www.image_gc, rw->www.anchorColor);
	XDrawRectangle(dpy, win, gc,
		       tp->x, tp->y,
		       tp->width, tp->height);
      }
      
      if (xi->depth == 1)
      {
	XSetForeground(dpy, gc, ImageForeground(ci->is));
	XSetBackground(dpy, gc, ImageBackground(ci->is));
      }
      XPutImage(dpy, win, gc, xi,
		0, fline,
		tp->x + fci->border_width,
		tp->y + fline + fci->border_width,
		tp->width, lline - fline + 1);
    }
  }

  return;
}

/*
 * AddImageRenderPart
 */
static void
AddImageRenderPart(rw, rp, width, height)
WWWWidget rw;
RenderPart *rp;
int width, height;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);

  ci->added = True;
  AddRenderPart(rw, rp, width + ci->border_width * 2,
		height + ci->border_width * 2);

  return;
}

/*
 * AddFakeRenderPart
 */
static void
AddFakeRenderPart(rw)
WWWWidget rw;
{
  RenderPart *rp;

  rp = CreateRenderPart(rw, NULL, NULL, NULL, NULL, NULL, 0);
  AddRenderPart(rw, rp, 0, 0);

  return;
}

/*
 * ProcessInlineDocument
 *
 * Returns False if the download should continue.
 */
static Boolean
ProcessInlineDocument(rp)
RenderPart *rp;
{
  InlineState *ci = (InlineState *)GetRenderState(rp);
  Document *d = ci->d;
  char *value;
  URLParts *up;
  int format;

  if ((value = MIMEGetFieldValue(d->mflist, "location", 0)) != NULL)
  {
    if ((up = ParseURL(value)) != NULL)
    {
      DestroyDocument(d);

      DestroyURLParts(ci->up);
      ci->up = up;

      up->reload = ci->rw->www.reload ? 1:0;
      LoadDocument((IOContext)ci->rw, up, HandleInlineDocument, ci);

      return(True);
    }
  }

  if ((value = MIMEGetFieldValue(d->mflist, "content-type", 0)) != NULL)
  {
    if (strcasecmp(value, "image/gif") == 0) format = IMAGE_GIF;
    else if (strcasecmp(value, "image/xbm") == 0) format = IMAGE_XBM;
    else if (strcasecmp(value, "image/x-xbitmap") == 0) format = IMAGE_XBM;
    else if (strcasecmp(value, "image/x-portable-anymap") == 0) format = IMAGE_PNM;
    else if (strcasecmp(value, "image/x-portable-bitmap") == 0) format = IMAGE_PNM;
    else if (strcasecmp(value, "image/x-portable-graymap") == 0) format = IMAGE_PNM;
    else if (strcasecmp(value, "image/x-portable-pixmap") == 0) format = IMAGE_PNM;
    else if (strcasecmp(value, "image/jpeg") == 0) format = IMAGE_JPEG;
    else return(True);

    if (ci->rw->www.ips == 0)
    {
      XWindowAttributes xwa;
      Pixel bg;
      Pixel fg;
      int status;

      status = XGetWindowAttributes(XtDisplay(ci->rw->www.child),
				    XtWindow(ci->rw->www.child),
				    &xwa);
      XtVaGetValues((Widget)ci->rw, XtNbackground, &bg, NULL);
      XtVaGetValues((Widget)ci->rw, XtNforeground, &fg, NULL);
      ci->rw->www.ips = ImagePInit(XtDisplay(ci->rw->www.child),
				   xwa.visual, xwa.depth,
				   xwa.colormap, bg, fg);
    }
    ci->is = ImageInit(ci->rw->www.ips, format, LineCallback, rp);
  }

  return(ci->is != 0 ? False:True);
}

/*
 * HandleInlineDocument
 */
static int
HandleInlineDocument(d, closure, status)
Document *d;
void *closure;
int status;
{
  RenderPart *rp = (RenderPart *)closure;
  InlineState *ci = (InlineState *)GetRenderState(rp);

  ci->d = d;

  if (status == DS_NEEDS_AUTH)
  {
    if (ci->up->username == NULL && FetchAuthInfo(ci->rw, ci->d, ci->up))
    {
      DestroyDocument(ci->d);
      ci->d = NULL;
      LoadDocument((IOContext)(ci->rw), ci->up, HandleInlineDocument, rp);
      return(-1);
    }
  }
  else if (status == DS_DOWNLOADING || status == DS_FINISHED)
  {
    if (ci->is == 0)
    {
      if (ProcessInlineDocument(rp))
      {
	AddFakeRenderPart(ci->rw);
	return(-1);
      }
    }
    if (d->len > 0) ImageAddData(ci->is, d->data, d->len);

    if (status == DS_FINISHED)
    {
      ImageEndData(ci->is);
      if (!ci->added) AddFakeRenderPart(ci->rw);
    }
  }
  else
  {
    /*
     * Don't add if the download was aborted
     */
    if (status != DS_ABORT && !ci->added) AddFakeRenderPart(ci->rw);
    return(-1);
  }

  return(0);
}

static void
DestroyInlineState(ci)
InlineState *ci;
{
  if (ci->up != NULL) DestroyURLParts(ci->up);
  if (ci->d != NULL) DestroyDocument(ci->d);
  if (ci->is != 0) ImageDestroy(ci->is);

  return;
}

static RenderPart *
CreateInlineState(rw, p, up)
WWWWidget rw;
MLTag *p;
URLParts *up;
{
  InlineState *ci;
  RenderPart *rp;

  rp = CreateRenderPart(rw, p,
			SetupInline,
			RenderInline, DestroyInlineState, NULL,
			sizeof(InlineState));
  ci = (InlineState *)GetRenderState(rp);
  ci->rw = rw;
  ci->d = NULL;
  ci->up = up;
  ci->mp = NULL;
  ci->next = NULL;
  ci->border_width = 1;
  ci->is = 0;
  ci->added = False;

  return(rp);
}

/*
 *
 * Public functions
 *
 */

/*
 * HandleImg
 */
void
HandleImg(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  char *url;
  URLParts *aup;
  URLParts *lup;
  URLParts *bup;
  InlineState *ci;
  RenderPart *rp;
  RenderPart *inline_rp;
#ifdef USE_HTML_WIDTH_HEIGHT
  unsigned int width, height;
#endif

  if (end != 0) return;

  if ((url = MLFindAttribute(p, "src")) == NULL) return;

  /*
   * Check if this image is already being downloaded.
   */
  for (rp = rw->www.rphead; rp != NULL; rp = rp->mnext)
  {
    if (rp->p != NULL && rp->p->id == TAG_IMG)
    {
      char *xurl;
      
      if ((xurl = MLFindAttribute(rp->p, "src")) != NULL &&
	  strcmp(xurl, url) == 0)
      {
	InlineState *xci;
	
	if ((xci = (InlineState *)GetRenderState(rp)) != NULL)
	{
	  inline_rp = CreateInlineState(rw, rp->p, NULL);
	  ci = (InlineState *)GetRenderState(inline_rp);

	  AddImageRenderPart(rw, inline_rp, rp->width, rp->height);
	
	  ci->next = xci->next;
	  xci->next = inline_rp;
	  ci->mp = rp;
	}

	return;
      }
    }
  }

  if ((lup = ParseURL(url)) == NULL) return;

  if ((bup = GetBaseURLParts(rw)) != NULL)
  {
    aup = MakeURLParts(lup, bup);
    DestroyURLParts(lup);
  }
  else aup = lup;

  inline_rp = CreateInlineState(rw, p, aup);
  ci = (InlineState *)GetRenderState(inline_rp);

  if (inline_rp->anchorPart != NULL) ci->border_width = 1;

#ifdef USE_HTML_WIDTH_HEIGHT
  if (GetWidthHeight(p, &width, &height) == 0)
  {
    AddImageRenderPart(rw, inline_rp, width, height);
  }
  else rw->www.delay = True;
#else
  rw->www.delay = True;
#endif
  
  url = MakeURL(aup, 0);
  XtCallCallbackList((Widget)rw, rw->www.downloadCallbacks, (XtPointer)url);
  free_mem(url);
  
  aup->reload = rw->www.reload ? 1:0;
  LoadDocument((IOContext)rw, aup, HandleInlineDocument, inline_rp);

  return;
}
