/*
 * io.c
 *
 * Copyright (C) 1994, 1995, John Kilburg (john@isri.unlv.edu).
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/types.h>
#include <errno.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <X11/IntrinsicP.h>

#include "WWWP.h"

static void TimeOutHandler _ArgProto((XtPointer, XtIntervalId *));
static void AddIOHandlers _ArgProto((IOInfo *, XtInputMask, void (*)()));
static void RemoveIOHandlers _ArgProto((IOInfo *));
static IOInfo *FindIOInfo _ArgProto((WWWWidget, int));

/*
 * FindIOInfo
 */
static IOInfo *
FindIOInfo(rw, s)
WWWWidget rw;
int s;
{
  IOInfo *c;

  for (c = rw->www.iilist; c != NULL; c = c->next)
  {
    if (c->s == s) return(c);
  }
  return(NULL);
}

/*
 * RemoveIOHandlers
 */
static void
RemoveIOHandlers(ii)
IOInfo *ii;
{
  if (ii->xid != 0) XtRemoveInput(ii->xid);
  ii->xid = 0;
  if (ii->iid != 0) XtRemoveTimeOut(ii->iid);
  ii->iid = 0;

  return;
}

/*
 * AddIOHandlers
 */
static void
AddIOHandlers(ii, iomask, iohandler)
IOInfo *ii;
XtInputMask iomask;
void (*iohandler)();
{
  ii->xid = XtAppAddInput(XtWidgetToApplicationContext((Widget)ii->rw),
			  ii->s, (XtPointer)iomask,
			  iohandler, (XtPointer)ii);
  
  ii->iid = XtAppAddTimeOut(XtWidgetToApplicationContext((Widget)ii->rw),
			    (unsigned long)ii->rw->www.ioTimeOutInterval,
			    TimeOutHandler, (XtPointer)ii);

  return;
}

/*
 * IOEnd
 */
void
IOEnd(ioc, s)
IOContext ioc;
int s;
{
  IOInfo *ii;
  IOQueue *iq;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "FinishIO IO error.\n");
    return;
  }

  RemoveIOHandlers(ii);

  if (ii->prev == NULL) rw->www.iilist = ii->next;
  else ii->prev->next = ii->next;
  if (ii->next != NULL) ii->next->prev = ii->prev;

  close(ii->s);

  free_mem(ii);

  rw->www.downloads--;
  if (rw->www.downloads < rw->www.maxDownloads)
  {
    if ((iq = rw->www.ioqhead) != NULL)
    {
      rw->www.ioqhead = rw->www.ioqhead->next;
      (*(iq->func))(iq->closure, DS_DOWNLOADING);
      free_mem(iq);
    }
  }

  return;
}

/*
 * IORegister
 *
 * Register a descriptor
 */
void
IORegister(ioc, s, cfunc, closure)
IOContext ioc;
int s;
void (*cfunc)();
void *closure;
{
  WWWWidget rw = (WWWWidget)ioc;
  IOInfo *ii;

  rw->www.downloads++;

  ii = (IOInfo *)alloc_mem(sizeof(IOInfo));
  ii->s = s;
  ii->func = NULL;
  ii->cfunc = cfunc;
  ii->closure = closure;

  ii->linelen = 0;

  ii->rdmax = 0;
  ii->rddata = NULL;
  ii->rdlen = 0;
  ii->rdmemsize = 0;

  ii->wri = 0;
  ii->wrdata = NULL;
  ii->wrlen = 0;

  ii->xid = 0;
  ii->rw = rw;
  ii->iid = 0;

  ii->prev = NULL;
  ii->next = rw->www.iilist;
  if (rw->www.iilist != NULL) rw->www.iilist->prev = ii;
  rw->www.iilist = ii;

  return;
}

/*
 * IOStart
 *
 * Indicate interest in doing IO
 */
void
IOStart(ioc, func, closure)
IOContext ioc;
void (*func)();
void *closure;
{
  WWWWidget rw = (WWWWidget)ioc;
  IOQueue *iq;

  if (rw->www.downloads >= rw->www.maxDownloads)
  {
    iq = (IOQueue *)alloc_mem(sizeof(IOQueue));
    iq->func = func;
    iq->closure = closure;
    iq->next = NULL;

    if (rw->www.ioqhead == NULL) rw->www.ioqhead = iq;
    else rw->www.ioqtail->next = iq;
    rw->www.ioqtail = iq;
  }
  else
  {
    (*func)(closure, DS_DOWNLOADING);
  }

  return;
}

/*
 * IOStop
 */
void
IOStop(rw)
WWWWidget rw;
{
  IOQueue *t;

  while (rw->www.ioqhead != NULL)
  {
    t = rw->www.ioqhead;
    rw->www.ioqhead = rw->www.ioqhead->next;
    (*(t->func))(t->closure, DS_ABORT);
    free_mem(t);
  }

  rw->www.ioqtail = NULL;
  rw->www.downloads = 0;

  while (rw->www.iilist != NULL)
  {
    (*(rw->www.iilist->cfunc))(rw->www.iilist->s,
			       rw->www.iilist->closure,
			       DS_ABORT);
  }

  return;
}

/*
 * IONotDone
 */
Boolean
IONotDone(rw)
WWWWidget rw;
{
  if (rw->www.iilist != NULL) return(True);
  return(False);
}

/*
 * WriteHandler
 */
static void
WriteHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  int rval;
  IOInfo *ii = (IOInfo *)cldata;

  rval = write(ii->s, (byte *)ii->wrdata + ii->wri, ii->wrlen - ii->wri);
  if (rval < 0)
  {
    if (errno != EWOULDBLOCK)
    {
      RemoveIOHandlers(ii);
      (*(ii->cfunc))(ii->s, ii->closure, DS_ERROR);
    }
  }
  else
  {
    ii->wri += rval;
    if (ii->wri == ii->wrlen)
    {
      RemoveIOHandlers(ii);
      (*(ii->func))(ii->s, ii->closure);
    }
  }

  return;
}

/*
 * ConnectHandler
 */
void
ConnectHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  IOInfo *ii = (IOInfo *)cldata;

  RemoveIOHandlers(ii);
  (*(ii->func))(ii->s, ii->closure);

  return;
}

/*
 * ReadLineHandler
 *
 * Ouch.  This severely sucks but makes life easier.
 */
static void
ReadLineHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  IOInfo *ii = (IOInfo *)cldata;
  int rval;
  
  rval = read(ii->s, ii->line + ii->linelen, 1);
  if (rval < 0)
  {
    if (errno != EWOULDBLOCK)
    {
      RemoveIOHandlers(ii);
      (*(ii->cfunc))(ii->s, ii->closure, DS_ERROR);
      return;
    }
  }
  else if (rval > 0)
  {
    if (ii->line[ii->linelen] == '\n')
    {
      ii->line[ii->linelen + 1] = '\0';
      RemoveIOHandlers(ii);
      (*(ii->func))(ii->s, ii->line, ii->closure);
      return;
    }
    ii->linelen++;
  }
  else
  {
    ii->line[ii->linelen] = '\0';
    RemoveIOHandlers(ii);
    (*(ii->func))(ii->s, ii->line, ii->closure);
    return;
  }

  return;
}

/*
 * ReadCharHandler
 */
static void
ReadCharHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  IOInfo *ii = (IOInfo *)cldata;
  int rval;
  unsigned char c;
  
  rval = read(ii->s, &c, 1);
  if (rval < 0)
  {
    if (errno != EWOULDBLOCK)
    {
      RemoveIOHandlers(ii);
      (*(ii->cfunc))(ii->s, ii->closure, DS_ERROR);
      return;
    }
  }
  else if (rval > 0)
  {
    RemoveIOHandlers(ii);
    (*(ii->func))(ii->s, &c, ii->closure);
  }

  return;
}

/*
 * ReadStreamHandler
 */
static void
ReadStreamHandler(cldata, netfd, xid)
XtPointer cldata;
int *netfd;
XtInputId *xid;
{
  IOInfo *ii = (IOInfo *)cldata;
  int rlen;
  int readsize;
  WWWWidget rw = ii->rw;

  /*
   * Figure out how much to read.  Try for ioQuantum unless there isn't
   * that much left to read.
   */
  if (ii->rdmax > 0)
  {
    if (rw->www.ioQuantum < ii->rdmax - ii->rdlen)
    {
      readsize = rw->www.ioQuantum;
    }
    else readsize = ii->rdmax - ii->rdlen;
  }
  else readsize = rw->www.ioQuantum;

  /*
   * Allocate buffer space for the read.  If there has been no allocation
   * then allocate the maximum if we know it so we don't have to allocate
   * memory again.  If we have already allocated some memory then realloc
   * more memory to accomodate the new read.
   *
   * Everything in multiples of BUFSIZ just because it might actually
   * make a difference.
   */
  if (ii->rddata == NULL)
  {
    ii->rdmemsize = (ii->rdmax / BUFSIZ + 1) * BUFSIZ;
    ii->rddata = alloc_mem(ii->rdmemsize);
  }
  else if (readsize > ii->rdmemsize - ii->rdlen)
  {
    ii->rdmemsize += (readsize / BUFSIZ + 1) * BUFSIZ;
    ii->rddata = realloc_mem(ii->rddata, ii->rdmemsize);
  }

  rlen = read(ii->s, (byte *)ii->rddata + ii->rdlen, readsize);
  if (rlen > 0)
  {
    ii->rdlen += rlen;

    /*
     * Callback periodically only when we know the maximum size.  This
     * prevents the address of the buffer being sent from changing due to
     * realloc()'s.  Means with some dumb servers we can't do on-the-fly
     * processing.
     */
    if (ii->rdmax > 0)
    {
      if (ii->rdlen < ii->rdmax)
      {
	RemoveIOHandlers(ii);
	(*(ii->func))(ii->s, ii->rddata, ii->rdlen, 0, ii->closure);
      }
      else
      {
	RemoveIOHandlers(ii);
	(*(ii->func))(ii->s, ii->rddata, ii->rdlen, 1, ii->closure);
      }
    }
  }
  else if (rlen < 0)
  {
    if (errno != EWOULDBLOCK)
    {
      RemoveIOHandlers(ii);
      (*(ii->cfunc))(ii->s, ii->closure, DS_ERROR);
    }
  }
  else
  {
    RemoveIOHandlers(ii);
    (*(ii->func))(ii->s, ii->rddata, ii->rdlen, 1, ii->closure);
  }

  return;
}

/*
 * TimeOutHandler
 */
static void
TimeOutHandler(closure, iid)
XtPointer closure;
XtIntervalId *iid;
{
  IOInfo *ii = (IOInfo *)closure;

  RemoveIOHandlers(ii);
  (*(ii->cfunc))(ii->s, ii->closure, DS_TIMEOUT);

  return;
}

/*
 * ReadLine
 */
void
IOReadLine(ioc, s, func)
IOContext ioc;
int s;
void (*func)();
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "ReadLine IO error.\n");
    return;
  }

  ii->func = func;
  ii->linelen = 0;
  AddIOHandlers(ii, XtInputReadMask, ReadLineHandler);

  return;
}

/*
 * ReadChar
 */
void
IOReadChar(ioc, s, func)
IOContext ioc;
int s;
void (*func)();
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "ReadChar IO error.\n");
    return;
  }

  ii->func = func;
  AddIOHandlers(ii, XtInputReadMask, ReadCharHandler);

  return;
}

/*
 * ReadStream
 */
void
IOReadStream(ioc, s, func)
IOContext ioc;
int s;
void (*func)();
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "ReadStream IO error.\n");
    return;
  }

  ii->func = func;
  AddIOHandlers(ii, XtInputReadMask, ReadStreamHandler);

  return;
}

/*
 * ConnectStream
 */
void
IOConnectStream(ioc, s, func)
IOContext ioc;
int s;
void (*func)();
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "ConnectStream IO error.\n");
    return;
  }

  ii->func = func;
  AddIOHandlers(ii, XtInputWriteMask, ConnectHandler);

  return;
}

/*
 * IOWriteStream
 */
void
IOWriteStream(ioc, s, t, tlen, func)
IOContext ioc;
int s;
void *t;
int tlen;
void (*func)();
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "WriteStream IO error.\n");
    return;
  }

  ii->wrdata = t;
  ii->wrlen = tlen;
  ii->wri = 0;
  ii->func = func;
  AddIOHandlers(ii, XtInputWriteMask, WriteHandler);

  return;
}

/*
 * IOResetStream
 */
void
IOResetStream(ioc, s, max)
IOContext ioc;
int s;
int max;
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "ResetStream IO error.\n");
    return;
  }

  if (ii->rddata != NULL)
  {
    free_mem(ii->rddata);
    ii->rddata = NULL;
  }

  ii->rdmax = max;
  ii->rdlen = 0;
  ii->rdmemsize = 0;

  ii->wri = 0;

  return;
}

/*
 * IOSetMax
 */
void
IOSetMax(ioc, s, max)
IOContext ioc;
int s;
int max;
{
  IOInfo *ii;
  WWWWidget rw = (WWWWidget)ioc;

  if ((ii = FindIOInfo(rw, s)) == NULL)
  {
    fprintf (stderr, "ResetStream IO error.\n");
    return;
  }

  ii->rdmax = max;

  return;
}
