/*
 * document.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "wwwio.h"

struct name
{
  char *name;
  void (*func)();
};

static struct name protocols[] = 
{
  { "ftp", ftp },
  { "http", http },
  { "gopher", gopher },
  { "gopherp", gopherp },
  { "file", file },
  { NULL, NULL },
};

static int LocalP _ArgProto((IOContext, URLParts *, int (*)(), void *));
static int ProxyP _ArgProto((IOContext, URLParts *, int (*)(), void *));
static int LocalProxyP _ArgProto((IOContext, URLParts *, int (*)(), void *));

/*
 * LocalProxyProtocol
 *
 */
static int
LocalProxyP(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  char *pproxy;
  char *pname;

  pname = (char *)alloc_mem(strlen(up->protocol) + strlen("_local") + 1);
  sprintf (pname, "%s_local", up->protocol);

  if ((pproxy = getenv(pname)) == NULL)
  {
    pproxy = NGetFromStringDB(pname);
    if (pproxy == NULL)
    {
      free_mem(pname);
      return(-1);
    }
  }

  http_local(ioc, up, pproxy, func, closure);

  free_mem(pname);

  return(0);
}

/*
 * ProxyP
 *
 * Patched heavily by Erik Corry <erik@kroete2.freinet.de>.
 */
static int
ProxyP(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  int plen, hlen;
  Document *d = NULL;
  URLParts *rup;
  char *pproxy = NULL;
  char *no_proxy;
  char *cp, *ph;
  char *pname;

  /*
   * file: URLs are never put through a proxy.
   */
  if(strcasecmp(up->protocol, "file") == 0)
  {
    return(-1);
  }

  /*
   * Hee, allow any weird protocol to be checked for proxy.
   */
  pname = (char *)alloc_mem(strlen(up->protocol) + strlen("_proxy") + 1);
  sprintf (pname, "%s_proxy", up->protocol);

  if ((pproxy = getenv(pname)) == NULL)
  {
    pproxy = NGetFromStringDB(pname);
    if (pproxy == NULL)
    {
      if ((pproxy = getenv("all_proxy")) == NULL)
      {
	pproxy = NGetFromStringDB("all_proxy");
	if (pproxy == NULL)
	{
	  free_mem(pname);
	  return(-1);
	}
      }
    }
  }
  pproxy = alloc_string(pproxy);
  free_mem(pname);

  /* 
   * Check for domains that are not allowed for proxy
   *
   * Default to not proxying localhost. A null no_proxy string
   * overrides this. TODO: get canonical hostname of current
   * host and exclude this by default, too.
   */
  if ((no_proxy = getenv("no_proxy")) == NULL
      && (no_proxy = NGetFromStringDB("no_proxy")) == NULL)
  {
    no_proxy = "localhost";
  }

  hlen = strlen(up->hostname);
  cp = no_proxy;
  while ((ph = mystrtok(cp, ',', &cp)) != NULL)
  {
    /*
     * Check whether the no_proxy variable either matches the
     * hostname exactly, or has the same ending. The ending must
     * be a domain name, not part of a domain name ie. no_proxy=
     * era.org does not match hostname=www.chimera.org.
     */
    plen = strlen(ph);
    if (plen != 0 &&
        plen <= hlen &&
        strcasecmp(ph, up->hostname + (hlen - plen)) == 0 &&
        (plen == hlen || up->hostname[hlen - plen - 1] == '.'
	 || up->hostname[hlen - plen] == '.'))
    {
      free_mem(pproxy);
      return(-1);
    }
  }
  
  /*
   * Try to get the proxy connection going.
   */
  if ((rup = ParseURL(pproxy)) != NULL)
  {
    char *msg = GetFromStringDB("absrurl");
    
    if (IsAbsoluteURL(rup))
    {
      http_proxy(ioc, rup, up, func, closure);
    }
    else
    {
      d = BuildDocument(msg, strlen(msg), "text/html", 1);
      (func)(d, closure, DS_FINISHED);
    }
    
    DestroyURLParts(rup);
  }
  else
  {
    char *msg = GetFromStringDB("invurl");
    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    (func)(d, closure, DS_FINISHED);
  }

  free_mem(pproxy);

  return(0);
}

/*
 * LocalPr
 *
 * Handle protocols that have been implemented in the chimera source.
 */
static int
LocalP(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  int i;

  for (i = 0; protocols[i].name != NULL; i++)
  {
    if (strcasecmp(protocols[i].name, up->protocol) == 0)
    {
      (protocols[i].func)(ioc, up, func, closure);
      return(0);
    }
  }

  return(-1);
}

/*
 * CreateDocument
 *
 * Allocates the structure used for documents.  Could do other
 * init stuff, too.  This function should always be called to create
 * a new document.  It provides initialization for the document which
 * is good if an old part of the program doesn't know about new
 * document attributes.
 */
Document *
CreateDocument()
{
  Document *d;

  d = (Document *)alloc_mem(sizeof(Document));
  d->data = NULL;
  d->len = 0;
  d->mflist = NULL;

  return(d);
}

/*
 * DestroyDocument
 */
void
DestroyDocument(d)
Document *d;
{
  MIMEField *m, *t;

  if (d->data != NULL) free_mem(d->data);

  for (m = d->mflist; m != NULL; )
  {
    t = m;
    m = m->next;
    MIMEDestroyField(t);
  }

  free_mem((char *)d);

  return;
}

/*
 * BuildDocument
 *
 * Do some of the grunge work to make a document.
 */
Document *
BuildDocument(data, len, content, doalloc)
void *data;
int len;
char *content;
int doalloc;
{
  Document *d;

  d = CreateDocument();
  if (doalloc && data != NULL)
  {
    d->data = alloc_mem(len);
    memcpy(d->data, data, len);
  }
  else
  {
    d->data = data;
  }
  d->len = (data == NULL ? 0:len);

  if (content != NULL) MIMEAddHeaderField(d, "Content-type", content);

  return(d);
}

/*
 * LoadDocument
 */
void
LoadDocument(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  if (!IsAbsoluteURL(up))
  {
    char *msg = GetFromStringDB("absurl");
    Document *d;

    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    (func)(d, closure, DS_FINISHED);
    return;
  }

  if (LocalProxyP(ioc, up, func, closure) == -1)
  {
    if (ProxyP(ioc, up, func, closure) == -1)
    {
      if (LocalP(ioc, up, func, closure) == -1)
      {
	char *msg = GetFromStringDB("badurl");
	Document *d;
	
	d = BuildDocument(msg, strlen(msg), "text/html", 1);
	(func)(d, closure, DS_FINISHED);
      }
    }
  }

  return;
}
